﻿// File:    $RCSfile: Base57.cs,v $
// Version: $Revision: 1.5 $
// Date:    $Date: 2019/09/09 11:51:39 $

namespace IQRFcode
{
  using System;
  using System.Text;

  /// <summary>
  /// Provides encoding and decoding an array of bytes into respectively from an alphanumeric string.
  /// Based on https://en.wikipedia.org/wiki/Base58 but uses only 57 characters as 57^11 ~= 2^64
  /// </summary>
  public static class Base57
  {
    /// <summary>
    /// Base (number of characters in the alphabet) used to encode.
    /// </summary>
    private const int base57 = 57;

    /// <summary>
    /// Alphabet used for alphanumeric coding.
    /// Contains digits and uppercase+lowercase letters except 0, I, O, l, u.
    /// </summary>
    private const string base57alphabet = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstvwxyz";

    /// <summary>
    /// Number of base57 character to encode full piece (8 bytes) .
    /// </summary>
    private const int maximumPiece57Length = 11;

    /// <summary>
    /// Array to convert from number of bytes into number of base57 to encode them.
    /// <para></para>
    /// base57chars base2bites  base2bitsRoundedDown
    /// 2           11.67       8
    /// 3           17.50       16
    /// 5           29.16       24
    /// 6           35.00       32
    /// 7           40.83       40
    /// 9           52.50       48
    /// 10          58.33       56
    /// 11          64.16       64
    /// </summary>
    private static readonly int[] piece57Lengths = { 0, 2, 3, 5, 6, 7, 9, 10, maximumPiece57Length };

    /// <summary>
    /// Encodes array of bytes into base57 character including check character.
    /// </summary>
    /// <param name="bytes">Bytes to encode.</param>
    /// <returns>Encoded bytes including check character.</returns>
    public static string EncodeBytes ( in byte[] bytes )
    {
      if ( bytes == null || bytes.Length == 0 )
        throw new ArgumentException( $"{nameof( Base57 )}.{nameof( EncodeBytes )}: Parameter is null or empty.", nameof( bytes ) );

      StringBuilder result = new StringBuilder();
      for ( int pieceIndex = 0; pieceIndex < bytes.Length; pieceIndex += sizeof( UInt64 ) )
      {
        int pieceLength = Math.Min( sizeof( UInt64 ), bytes.Length - pieceIndex );

        UInt64 piece = 0;
        for ( int inPieceIndex = 0; inPieceIndex < pieceLength; inPieceIndex++ )
          piece = ( piece << 8 ) | bytes[pieceIndex + inPieceIndex];

        for ( int piece57Length = piece57Lengths[pieceLength]; piece57Length != 0; piece57Length-- )
        {
          result.Append( base57alphabet[(int)( piece % base57 )] );
          piece /= base57;
        }
      }

      return result.Append( CheckChar( result.ToString() ) ).ToString();
    }

    /// <summary>
    /// Decodes base57 + check character string into bytes.
    /// </summary>
    /// <param name="text">Encoded bytes.</param>
    /// <returns>Decoded array of bytes.</returns>
    public static byte[] DecodeText ( string text )
    {
      if ( string.IsNullOrEmpty( text ) )
        throw new ArgumentException( $"{nameof( Base57 )}.{nameof( DecodeText )}: Parameter is null or empty.", nameof( text ) );

      char checkChar = text[text.Length - 1];
      text = text.Substring( 0, text.Length - 1 );
      if ( checkChar != CheckChar( text ) )
        throw new ArgumentException( $"{nameof( Base57 )}.{nameof( DecodeText )}: Parameter ends by an incorrect check character.", nameof( text ) );

      int lastPiece57Length = text.Length % maximumPiece57Length;
      int lastPieceLength = Array.IndexOf( piece57Lengths, lastPiece57Length );
      if ( lastPieceLength == -1 )
        throw new ArgumentException( $"{nameof( Base57 )}.{nameof( DecodeText )}: Parameter has incorrect length {text.Length}.", nameof( text ) );

      byte[] result = new byte[sizeof( UInt64 ) * ( text.Length / maximumPiece57Length ) + lastPieceLength];
      for ( int piece57Index = 0, pieceIndex = 0; piece57Index < text.Length; piece57Index += maximumPiece57Length, pieceIndex += sizeof( UInt64 ) )
      {
        int piece57Length = Math.Min( maximumPiece57Length, text.Length - piece57Index );

        UInt64 piece = 0;
        for ( int inPiece57Index = piece57Length; --inPiece57Index >= 0; )
        {
          char char57 = text[piece57Index + inPiece57Index];
          int alphabet57Index = base57alphabet.IndexOf( char57 );
          if ( alphabet57Index == -1 )
            throw new ArgumentException( $"{nameof( Base57 )}.{nameof( DecodeText )}: Parameter contains an incorrect character '{char57}' at position {piece57Index + inPiece57Index}.", nameof( text ) );

          piece = piece * base57 + (UInt64)alphabet57Index;
        }

        for ( int pieceLength = Array.IndexOf( piece57Lengths, piece57Length ); --pieceLength >= 0; )
        {
          result[pieceIndex + pieceLength] = (byte)( piece & 0xFF );
          piece >>= 8;
        }
      }

      return result;
    }

    /// <summary>
    /// Computes check character for the given string using https://en.wikipedia.org/wiki/Luhn_mod_N_algorithm 
    /// </summary>
    /// <param name="text">String to return the check character for.</param>
    /// <returns>Check character.</returns>
    private static char CheckChar ( in string text )
    {
      if ( string.IsNullOrEmpty( text ) )
        throw new ArgumentException( $"{nameof( Base57 )}.{nameof( CheckChar )}: Parameter is null or empty.", nameof( text ) );

      int sum = 0;
      for ( int index = text.Length - 1, factor = 2; index >= 0; index--, factor = ( 1 + 2 ) - factor )
      {
        int addend = factor * base57alphabet.IndexOf( text[index] );
        sum += addend / base57 + ( addend % base57 );
      }

      return base57alphabet[( base57 - ( sum % base57 ) ) % base57];
    }
  }
}
