﻿// File:    $RCSfile: IQRFcode.cs,v $
// Version: $Revision: 1.15 $
// Date:    $Date: 2019/09/09 11:51:39 $

namespace IQRFcode
{
  using System;
  using System.Collections.Generic;
  using System.Text;
  using static System.Console;

  /// <summary>
  /// Command line console application used to encode/decode IQRF Codes.
  /// </summary>
  internal class IQRFcode
  {
    /// <summary>
    /// Tags to mark important IQRF data objects.
    /// </summary>
    private enum Tags : byte
    {
      /// <summary>
      /// Marks end of the nibble stream.
      /// </summary>
      End = 0x0,
      MID = 0x1,
      IBK = 0x2,
      HWPID = 0x3,
      BondChannel = 0x4,
      Nop = 0x05,
      DataBlock = 0x06,
      Text = 0x07,

      /// <summary>
      /// For future tags count expansion.
      /// </summary>
      TAG_RESERVED = 0xF
    }

    /// <summary>
    /// Program entry point.
    /// </summary>
    /// <param name="args">Command-line arguments.</param>
    /// <returns>Application result code.</returns>
    /// <example>
    /// Command-line examples:
    ///  encode -MID:12345678 -HwpId:AABB -IBK:00112233445566778899AABBCCDDEEFF
    ///  encode -Nop: -MID:81111E95 -Nop: -IBK:AB79EA9000000000000000000000006C -Nop: -HWPID:ABCD -s:
    ///  encode -s: -Nop: -MID:81111E95 -Nop: -IBK:AB79EA90CEB0A218733A13FD54176C02 -Nop: -HwpId:0002 -Nop: -DataBlock:000102EE -Nop: "-Text:A Z" -DataBlock:11223344 -Nop: "-Text:a z"
    ///  decode -Code:DSjAtrBCHc7kDjkR4MXzC438XUwR5ByNXM
    ///  decode -Code:d1MNeMQ3yG5jK68QTgw9goJBQaem5bmy4gb5SSLkVAU1LviFs991
    /// </example>
    public static int Main ( string[] args )
    {
      try
      {
        int argsIndex = 0;

        string nextArg ( out string option, out string param, in bool paramPresent )
        {
          if ( argsIndex >= args.Length )
            return option = param = null;

          string result = args[argsIndex++];
          int delimIndex = result.IndexOf( ':' );
          if ( delimIndex == -1 )
          {
            if ( paramPresent )
              throw new Exception( $"Data for option {result} is missing." );

            option = result;
            param = null;
          }
          else
          {
            option = result.Substring( 0, delimIndex );

            if ( !paramPresent )
              throw new Exception( $"Data for option {option} is present." );

            param = delimIndex + 1 < result.Length ? result.Substring( delimIndex + 1 ) : "";
          }

          return result;
        }

        string verb = nextArg( out string _, out string _, false );
        if ( verb == null )
          throw new Exception( "IQRFcode V0.01" );

        switch ( verb.ToUpper() )
        {
          case "ENCODE":
          {
            bool outputNibbles = false;
            NibbleWriter writer = new NibbleWriter();

            bool ibk = false, mid = false, hwpid = false, bondChannel = false;
            while ( nextArg( out string option, out string param, true ) != null )
            {
              switch ( option.ToUpper() )
              {
                case "-MID":
                  if ( mid )
                    throw new Exception( $"MID was already specified." );
                  mid = true;

                  writer.WriteNibbleAndBytes( (byte)Tags.MID, ParseHexBytes( param, 4 ) );
                  break;

                case "-IBK":
                  if ( ibk )
                    throw new Exception( $"IBK was already specified." );
                  ibk = true;

                  writer.WriteNibbleAndBytes( (byte)Tags.IBK, ParseHexBytes( param, 16 ) );
                  break;

                case "-HWPID":
                  if ( hwpid )
                    throw new Exception( "HWPID was already specified." );
                  hwpid = true;

                  writer.WriteNibbleAndBytes( (byte)Tags.HWPID, ParseHexBytes( param, 2 ) );
                  break;

                case "-BONDCHAN":
                  if ( bondChannel )
                    throw new Exception( "Bonding channel was already specified." );
                  Error.WriteLine( "Warning: obsolete parameter BONDCHAN used." );
                  bondChannel = true;

                  if ( !byte.TryParse( param, out byte bondingChannel ) )
                    throw new Exception( "Invalid bonding channel value." );

                  writer.WriteNibble( (byte)Tags.BondChannel );
                  writer.WriteByte( bondingChannel );
                  break;

                case "-NOP":
                  writer.WriteNibble( (byte)Tags.Nop );
                  break;

                case "-S":
                  outputNibbles = true;
                  break;

                case "-DATABLOCK":
                {
                  writer.WriteNibble( (byte)Tags.DataBlock );
                  byte[] datablock = ParseHexBytes( param, -1 );
                  writer.WriteByte( (byte)datablock.Length );
                  writer.WriteBytes( datablock );
                  break;
                }

                case "-TEXT":
                {
                  writer.WriteNibbleAndBytes( (byte)Tags.Text, Encoding.UTF8.GetBytes( param ) );
                  writer.WriteByte( 0 );
                  break;
                }

                default:
                  throw new Exception( $"Unknown option {option}." );
              }
            }

            writer.WriteNibble( (byte)Tags.End );

            byte[] nibbles = writer.Close();

            if ( outputNibbles )
              OutputNibbles( nibbles );
            WriteLine( Base57.EncodeBytes( nibbles ) );
            break;
          }

          case "DECODE":
          {
            string iqrfCode = null;
            while ( nextArg( out string option, out string param, true ) != null )
            {
              switch ( option.ToUpper() )
              {
                case "-CODE":
                  if ( iqrfCode != null )
                    throw new Exception( $"IQRF Code was already specified." );

                  iqrfCode = param;
                  break;

                default:
                  throw new Exception( $"Unknown option {option}." );
              }
            }

            NibbleReader reader = new NibbleReader( Base57.DecodeText( iqrfCode ?? throw new Exception( $"IQRF Code to decode was not specified." ) ) );
            for ( Tags tag; ( tag = (Tags)reader.ReadNibble() ) != Tags.End; )
              switch ( tag )
              {
                case Tags.MID:
                  readAndWriteHexBytes( "-MID:", 4 );
                  break;

                case Tags.IBK:
                  readAndWriteHexBytes( "-IBK:", 16 );
                  break;

                case Tags.HWPID:
                  readAndWriteHexBytes( "-HWPID:", 2 );
                  break;

                case Tags.BondChannel:
                  Write( $"-BondChan:{reader.ReadByte()} " );
                  break;

                case Tags.Nop:
                  Write( "-Nop: " );
                  break;

                case Tags.DataBlock:
                  readAndWriteHexBytes( "-DataBlock:", reader.ReadByte() );
                  break;

                case Tags.Text:
                {
                  List<byte> utf8bytes = new List<byte>();
                  for ( byte @byte; 0 != ( @byte = reader.ReadByte() ); )
                    utf8bytes.Add( @byte );
                  Write( $"\"-Text:{Encoding.UTF8.GetString( utf8bytes.ToArray(), 0, utf8bytes.Count )}\" " );
                  break;
                }

                default:
                  throw new Exception( "Unknown tag value." );
              }

            void readAndWriteHexBytes ( string label, int length )
            {
              Write( label );
              do
              {
                Write( $"{reader.ReadByte():X02}" );
              }
              while ( --length != 0 );
              Write( ' ' );
            }

            WriteLine();
            break;
          }

          default:
            throw new Exception( $"Unknown verb {verb}." );
        }

        return 0;
      }
#pragma warning disable CA1031 // Do not catch general exception types
      catch ( Exception ex )
      {
        Error.WriteLine( $"\r\n{ex.Message}" );
        Help();
        return 1;
      }
#pragma warning restore CA1031 // Do not catch general exception types
    }

    private static byte[] ParseHexBytes ( in string hexBytes, in int length )
    {
      if ( length != -1 && hexBytes.Length != length * 2 )
        throw new Exception( $"Hexadecimal bytes {hexBytes} are not {length * 2} character long." );

      if ( hexBytes.Length % 2 != 0 )
        throw new Exception( $"Hexadecimal bytes {hexBytes} length is not even." );

      List<byte> result = new List<byte>();
      for ( int index = 0; length != -1 ? index < length : index * 2 < hexBytes.Length; index++ )
      {
        if ( !byte.TryParse( hexBytes.Substring( index * 2, 2 ), System.Globalization.NumberStyles.HexNumber, null, out byte @byte ) )
          throw new Exception( $"Invalid hexadecimal character at {hexBytes}." );

        result.Add( @byte );
      }

      if ( result.Count < 1 || result.Count > byte.MaxValue )
        throw new Exception( $"Hexadecimal bytes {hexBytes} length {result.Count} is invalid." );

      return result.ToArray();
    }

    private static void OutputNibbles ( in byte[] nibbles )
    {
      Write( "Nibbles:" );
      foreach ( byte nibble in nibbles )
        Write( $" {nibble:X02}" );
      WriteLine();
    }

    private static void Help ()
    {
      Error.WriteLine(
        "\r\nUsage:\r\n" +
        "IQRFcode encode [options]\r\n" +
        "\t-MID:XxXxXxXx\t\tSets 8 bytes hexadecimal MID.\r\n" +
        "\t-IBK:Xx...Xx\t\tSets 16 bytes hexadecimal IBK.\r\n" +
        "\t-HWPID:XxXx\t\tSets 4 bytes hexadecimal HWPID.\r\n" +
        "\t-BondChan:X\t\t[obsolete] Sets decimal bonding channel.\r\n" +
        "\t-Nop:\t\t\tNo operation IQRF Value.\r\n" +
        "\t-DataBlock:Xx...Xx\tSets data block bytes.\r\n" +
        "\t-Text:string\t\tSets the text value.\r\n" +
        "\t-s:\t\t\tOutput the nibble stream.\r\n" +
        "IQRFcode decode [options]\r\n" +
        "\t-Code:Ab...yZ\t\tDecodes provided IQRF Code."
        );
    }
  }
}
