﻿// File:    $RCSfile: Nibbles.cs,v $
// Version: $Revision: 1.4 $
// Date:    $Date: 2019/09/09 11:51:39 $

namespace IQRFcode
{
  using System;
  using System.Collections.Generic;

  /// <summary>
  /// Nibble "stream" writer.
  /// </summary>
  public class NibbleWriter
  {
    private readonly List<byte> stream = new List<byte>();
    private int bitsWritten;
    private byte byte2Write;

    /// <summary>
    /// Writes one nibble.
    /// </summary>
    /// <param name="nibble">Nibble to write.</param>
    public void WriteNibble ( in byte nibble )
    {
      if ( nibble > 0x0F )
        throw new ArgumentException( $"{nameof( NibbleWriter )}.{nameof( WriteNibble )}: Parameter value is not nibble", nameof( nibble ) );

      byte2Write |= (byte)( nibble << bitsWritten );
      bitsWritten += 4;
      if ( bitsWritten >= 8 )
      {
        bitsWritten -= 8;
        stream.Add( byte2Write );
        byte2Write = 0;
      }
    }

    /// <summary>
    /// Writes one byte.
    /// </summary>
    /// <param name="byte">Byte to write.</param>
    public void WriteByte ( in byte @byte )
    {
      WriteNibble( (byte)( @byte & 0x0F ) );
      WriteNibble( (byte)( @byte >> 4 ) );
    }

    /// <summary>
    /// Writes array bytes.
    /// </summary>
    /// <param name="bytes">Bytes to write.</param>
    public void WriteBytes ( in byte[] bytes )
    {
      if ( bytes != null )
      {
        if ( bytes.Length == 0 )
          throw new ArgumentException( $"{nameof( NibbleWriter )}.{nameof( WriteBytes )}: Parameter is empty.", nameof( bytes ) );

        foreach ( byte @byte in bytes )
          WriteByte( @byte );
      }
    }

    /// <summary>
    /// Writes one nibble and array of bytes.
    /// </summary>
    /// <param name="nibble">Nibble to write.</param>
    /// <param name="bytes">Bytes to write.</param>
    public void WriteNibbleAndBytes ( in byte nibble, in byte[] bytes )
    {
      WriteNibble( nibble );
      WriteBytes( bytes );
    }

    /// <summary>
    /// Closes nibble writer and returns written bytes. No more data can be written then.
    /// </summary>
    /// <returns>Written bytes.</returns>
    public byte[] Close ()
    {
      if ( bitsWritten != 0 )
        stream.Add( byte2Write );

      return stream.ToArray();
    }
  }

  /// <summary>
  /// Nibble "stream" reader.
  /// </summary>
  public class NibbleReader
  {
    private readonly byte[] stream;
    private int nibbleIndex = 0;

    /// <summary>
    /// Initializes nibble reader using bytes array.
    /// </summary>
    /// <param name="bytes">Bytes to initialize reader.</param>
    public NibbleReader ( in byte[] bytes ) => stream = bytes;

    /// <summary>
    /// Reads one nibble.
    /// </summary>
    /// <returns>Read nibble.</returns>
    public byte ReadNibble ()
    {
      int streamIndex = nibbleIndex / 2;
      if ( streamIndex >= stream.Length )
        throw new Exception( "No more data available to read." );

      byte @byte = stream[streamIndex];
      if ( nibbleIndex++ % 2 == 1 )
        @byte >>= 4;
      return (byte)( @byte & 0x0F );
    }

    /// <summary>
    /// Reads one byte.
    /// </summary>
    /// <returns>Read byte.</returns>
    public byte ReadByte () => (byte)( ReadNibble() | ( ReadNibble() << 4 ) );
  }
}
